@@ -1,26 +1,19 @@
 # Mrs
 # Copyright 2008-2012 Brigham Young University
 #
-# This file is part of Mrs.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Mrs is free software: you can redistribute it and/or modify it under the
-# terms of the GNU General Public License as published by the Free Software
-# Foundation, either version 3 of the License, or (at your option) any later
-# version.
+#     http://www.apache.org/licenses/LICENSE-2.0
 #
-# Mrs is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# Mrs.  If not, see <http://www.gnu.org/licenses/>.
-#
-# Inquiries regarding any further use of Mrs, please contact the Copyright
-# Licensing Office, Brigham Young University, 3760 HBLL, Provo, UT 84602,
-# (801) 422-9339 or 422-3821, e-mail copyright@byu.edu.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
-from __future__ import division
+from __future__ import division, print_function
 
 import multiprocessing
 import os
@@ -99,6 +92,7 @@ class Job(object):
         if outdir:
             util.try_makedirs(outdir)
 
+        self._set_serializers(None, kwds)
         ds = datasets.LocalData(itr, splits, dir=outdir, parter=parter,
                 permanent=permanent, **kwds)
         if self._url_converter:
@@ -127,16 +121,16 @@ class Job(object):
         else:
             permanent = False
 
-        if not parter:
+        if parter is None:
             parter = self.default_partition
+        part_name, _ = self._named_attr(parter)
 
-        map_name = self._registry[mapper]
+        map_name, mapper = self._named_attr(mapper)
         self._set_serializers(mapper, kwds)
-        if combiner:
-            combine_name = self._registry[combiner]
+        if combiner is not None:
+            combine_name, _ = self._named_attr(combiner)
         else:
             combine_name = ''
-        part_name = self._registry[parter]
 
         op = tasks.MapOperation(map_name, combine_name, part_name)
         ds = computed_data.ComputedData(op, input, splits=splits, dir=outdir,
@@ -163,12 +157,12 @@ class Job(object):
         else:
             permanent = False
 
-        if not parter:
+        if parter is None:
             parter = self.default_partition
+        part_name, _ = self._named_attr(parter)
 
-        reduce_name = self._registry[reducer]
-        self._set_serializers(reducer, kwds)
-        part_name = self._registry[parter]
+        reduce_name, reducer = self._named_attr(reducer)
+        self._set_serializers(reducer, kwds, input.serializers)
 
         op = tasks.ReduceOperation(reduce_name, part_name)
         ds = computed_data.ComputedData(op, input, splits=splits, dir=outdir,
@@ -195,14 +189,14 @@ class Job(object):
         if not parter:
             parter = self.default_partition
 
-        reduce_name = self._registry[reducer]
-        map_name = self._registry[mapper]
+        reduce_name, reducer = self._named_attr(reducer)
+        map_name, mapper = self._named_attr(mapper)
         self._set_serializers(mapper, kwds)
-        if combiner:
-            combine_name = self._registry[combiner]
+        if combiner is not None:
+            combine_name, _ = self._named_attr(combiner)
         else:
             combine_name = ''
-        part_name = self._registry[parter]
+        part_name, _ = self._named_attr(parter)
 
         op = tasks.ReduceMapOperation(reduce_name, map_name, combine_name,
                 part_name)
@@ -216,28 +210,55 @@ class Job(object):
         """Reports the progress (fraction complete) of the given dataset."""
         return self._manager.progress(dataset)
 
-    def _set_serializers(self, f, kwds):
+    def _set_serializers(self, f, kwds, fallback_serializers=None):
         """Add any serializers specified on the given function to kwds."""
+
         if 'key_serializer' in kwds:
-            key_s_name = kwds['key_serializer']
+            key_s = kwds['key_serializer']
             del kwds['key_serializer']
         elif hasattr(f, 'key_serializer'):
-            key_s_name = f.key_serializer
+            key_s = f.key_serializer
+        elif fallback_serializers is not None:
+            key_s = fallback_serializers.key_s
         else:
-            key_s_name = None
+            key_s = None
+
+        if isinstance(key_s, str):
+            key_s_name = key_s
+            key_s = getattr(self._program, key_s)
+        elif key_s is None:
+            key_s_name = ''
+        else:
+            key_s_name = self._registry[key_s]
 
         if 'value_serializer' in kwds:
-            value_s_name = kwds['value_serializer']
+            value_s = kwds['value_serializer']
             del kwds['value_serializer']
         elif hasattr(f, 'value_serializer'):
-            value_s_name = f.value_serializer
+            value_s = f.value_serializer
+        elif fallback_serializers is not None:
+            value_s = fallback_serializers.value_s
         else:
-            value_s_name = None
+            value_s = None
 
-        names = key_s_name, value_s_name
-        serializers = Serializers.from_names(names, self._program)
+        if isinstance(value_s, str):
+            value_s_name = value_s
+            value_s = getattr(self._program, value_s)
+        elif value_s is None:
+            value_s_name = ''
+        else:
+            value_s_name = self._registry[value_s]
+
+        serializers = Serializers(key_s, key_s_name, value_s, value_s_name)
         kwds['serializers'] = serializers
 
+    def _named_attr(self, value):
+        if isinstance(value, str):
+            return value, getattr(self._program, value)
+        else:
+            return self._registry[value], value
+
+
 
 def job_process(program_class, opts, args, default_dir, pipe,
         quit_pipe, use_bucket_server):
