@@ -1,24 +1,17 @@
 # Mrs
 # Copyright 2008-2012 Brigham Young University
 #
-# This file is part of Mrs.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Mrs is free software: you can redistribute it and/or modify it under the
-# terms of the GNU General Public License as published by the Free Software
-# Foundation, either version 3 of the License, or (at your option) any later
-# version.
+#     http://www.apache.org/licenses/LICENSE-2.0
 #
-# Mrs is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# Mrs.  If not, see <http://www.gnu.org/licenses/>.
-#
-# Inquiries regarding any further use of Mrs, please contact the Copyright
-# Licensing Office, Brigham Young University, 3760 HBLL, Provo, UT 84602,
-# (801) 422-9339 or 422-3821, e-mail copyright@byu.edu.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 """Mrs Master"""
 
@@ -75,7 +68,6 @@ class MasterRunner(runner.TaskRunner):
         self.slaves = None
         self.idle_slaves = IdleSlaves()
         self.dead_slaves = set()
-        self.current_assignments = set()
         self.result_maps = {}
 
         self.rpc_interface = None
@@ -125,10 +117,11 @@ class MasterRunner(runner.TaskRunner):
 
     def maintain_chore_queue(self):
         """Maintains the chore_queue and returns the timeout value for poll."""
-        timeleft = self.chore_queue.time_to_reschedule()
+        chore_queue = self.chore_queue
+        timeleft = chore_queue.time_to_reschedule()
         if (timeleft is not None) and (timeleft <= 0):
-            self.chore_queue.reschedule()
-            timeleft = self.chore_queue.time_to_reschedule()
+            chore_queue.reschedule()
+            timeleft = chore_queue.time_to_reschedule()
         return timeleft
 
     def read_sched_pipe(self):
@@ -169,13 +162,11 @@ class MasterRunner(runner.TaskRunner):
                 # Dataset already deleted, so this source should be removed.
                 self.remove_sources(dataset_id, (slave, source), delete=True)
 
-            if (dataset_id, source) in self.current_assignments:
-                # Note: if this is the last task in the dataset, this will wake
-                # up datasets.  Thus this happens _after_ slaves are added to
-                # the idle_slaves set.
-                self.task_done(dataset_id, source, urls)
-                self.current_assignments.remove((dataset_id, source))
-            else:
+            # Note: if this is the last task in the dataset, this will wake
+            # up datasets.  Thus this happens _after_ slaves are added to
+            # the idle_slaves set.
+            success = self.task_done(dataset_id, source, urls)
+            if not success:
                 logger.info('Ignoring a redundant result (%s, %s).' %
                         (dataset_id, source))
 
@@ -226,7 +217,6 @@ class MasterRunner(runner.TaskRunner):
             slave.prepare_assignment(next, self.datasets)
             chore_item = slave.send_assignment, ()
             chore_list.append(chore_item)
-            self.current_assignments.add(next)
         self.chore_queue.do_many(chore_list)
 
     def available_workers(self):
@@ -268,10 +258,8 @@ class MasterRunner(runner.TaskRunner):
 
     def debug_status(self):
         super(MasterRunner, self).debug_status()
-        print('Current assignments:', (', '.join('(%s, %s)' % a
-                for a in self.current_assignments)), file=sys.stderr)
         print('Idle slaves:', (', '.join(str(slave.id)
-                for slave in self.idle_slaves)), file=sys.stderr)
+                for slave in self.idle_slaves._all_slaves)), file=sys.stderr)
         print('Dead slaves:', (', '.join(str(slave.id)
                 for slave in self.dead_slaves)), file=sys.stderr)
 
@@ -467,16 +455,35 @@ class RemoteSlave(object):
         """
         return self._assignment
 
-    def set_assignment(self, old_assignment, new_assignment):
-        """Sets the assignment to new_assignment if the old_assignment matches.
+    def set_assignment(self, new_assignment):
+        """Sets the assignment to new_assignment.
+
+        Assumes that the assignment is currently None.  Returns True if the
+        operation succeeds.
         """
         with self._assignment_lock:
-            if self._assignment == old_assignment:
+            if self._assignment is None:
                 self._assignment = new_assignment
                 return True
             else:
                 return False
 
+    def clear_assignment(self, old_assignment):
+        """Sets the assignment to None.
+
+        Assumes that the assignment is currently `old_assignment` or `None`.
+        Returns True if the old assignment was set or False if it was None.
+        """
+        with self._assignment_lock:
+            if self._assignment == old_assignment:
+                self._assignment = None
+                return True
+            elif self._assignment is None:
+                return False
+            else:
+                logger.error("clear_assignment called with an invalid argument")
+                return False
+
     def prepare_assignment(self, assignment, datasets):
         """Sets up an RPC request to make the slave work on the assignment.
 
@@ -484,7 +491,7 @@ class RemoteSlave(object):
         happen until `send_assignment` is subsequently called.  This is the
         responsibility of the caller.
         """
-        success = self.set_assignment(None, assignment)
+        success = self.set_assignment(assignment)
         assert success
         dataset_id, task_index = assignment
 
@@ -512,19 +519,15 @@ class RemoteSlave(object):
             try:
                 success = self._rpc_func(*self._rpc_args)
             except Fault as f:
-                logger.error('Fault in RPC call to slave %s: %s'
-                        % (self.id, f.faultString))
+                logger.error('Fault in RPC to slave %s: %s' %
+                        (self.id, f.faultString))
                 success = False
             except ProtocolError as e:
-                logger.error('Protocol error in RPC call to slave %s: %s'
-                        % (self.id, e.errmsg))
+                logger.error('Protocol error in RPC to slave %s: %s' %
+                        (self.id, e.errmsg))
                 success = False
-            except socket.timeout:
-                logger.error('Timeout in RPC call to slave %s' % self.id)
-                success = False
-            except socket.error as e:
-                logger.error('Socket error in RPC call to slave %s: %s'
-                        % (self.id, e.args[1]))
+            except http.ConnectionFailed:
+                logger.error('Connection failed in RPC to slave %s' % self.id)
                 success = False
 
             if success:
@@ -557,12 +560,9 @@ class RemoteSlave(object):
                 logger.error('Protocol error in remove call to slave %s: %s'
                         % (self.id, e.errmsg))
                 success = False
-            except socket.timeout:
-                logger.error('Timeout in remove call to slave %s' % self.id)
-                success = False
-            except socket.error as e:
-                logger.error('Socket error in remove call to slave %s: %s'
-                        % (self.id, e.args[1]))
+            except http.ConnectionFailed:
+                logger.error('Connection failed in remove call to slave %s' %
+                        self.id)
                 success = False
 
             if success:
@@ -649,12 +649,8 @@ class RemoteSlave(object):
             logger.error('Protocol error in ping to slave %s: %s'
                     % (self.id, e.errmsg))
             success = False
-        except socket.timeout:
-            logger.error('Timeout in ping to slave %s' % self.id)
-            success = False
-        except socket.error as e:
-            logger.error('Socket error in ping to slave %s: %s'
-                    % (self.id, e.args[1]))
+        except http.ConnectionFailed:
+            logger.error('Connection failed in ping to slave %s' % self.id)
             success = False
         finally:
             self._rpc_lock.release()
@@ -694,11 +690,8 @@ class RemoteSlave(object):
             except ProtocolError as e:
                 logger.error('Protocol error in exit to slave %s: %s'
                         % (self.id, e.errmsg))
-            except socket.timeout:
-                logger.error('Timeout in exit to slave %s' % self.id)
-            except socket.error as e:
-                logger.error('Socket error in exit to slave %s: %s'
-                        % (self.id, e.args[1]))
+            except http.ConnectionFailed:
+                logger.error('Connection failed in exit to slave %s' % self.id)
             self._state = 'exited'
             self._rpc = None
 
@@ -770,19 +763,33 @@ class Slaves(object):
 
         self.trigger_sched()
 
-    def slave_result(self, slave, dataset_id, source, urls):
-        success = slave.set_assignment((dataset_id, source), None)
-        assert success
-        self._results.append((slave, dataset_id, source, urls))
-        self._changed_slaves.append(slave)
-        self.trigger_sched()
+    def slave_result(self, slave, dataset_id, task_index, urls):
+        """Called when a slave reports a successfully completed assignment.
+
+        Note that in the case of retried timeouts, this may be called multiple
+        times.
+        """
+        success = slave.clear_assignment((dataset_id, task_index))
+        if success:
+            self._results.append((slave, dataset_id, task_index, urls))
+            self._changed_slaves.append(slave)
+            self.trigger_sched()
+        else:
+            logger.error("Ignoring a possibly duplicate slave_result call.")
 
     def slave_failed(self, slave, dataset_id, task_index):
-        success = slave.set_assignment((dataset_id, task_index), None)
-        assert success
-        self._failed_tasks.append((dataset_id, task_index))
-        self._changed_slaves.append(slave)
-        self.trigger_sched()
+        """Called when a slave reports a failed assignment.
+
+        Note that in the case of retried timeouts, this may be called multiple
+        times.
+        """
+        success = slave.clear_assignment((dataset_id, task_index))
+        if success:
+            self._failed_tasks.append((dataset_id, task_index))
+            self._changed_slaves.append(slave)
+            self.trigger_sched()
+        else:
+            logger.error("Ignoring a possibly duplicate slave_failed call.")
 
     def slave_dead(self, slave):
         self._changed_slaves.append(slave)
@@ -927,11 +934,19 @@ class IdleSlaves(object):
         if host_size > self._max_count:
             self._max_count = host_size
 
+    def _consistency_check(self):
+        assert self._max_count >= 0
+        max_count = 0
+        for count in self._counter.keys():
+            if self._counter[count] and count > max_count:
+                max_count = count
+        assert self._max_count == max_count
+
     def _remove_from_counter(self, host, host_size):
         if host_size != 0:
             counter_set = self._counter[host_size]
             counter_set.remove(host)
-            if not counter_set:
+            if self._max_count == host_size and not counter_set:
                 self._max_count -= 1
 
     def __nonzero__(self):
@@ -943,4 +958,7 @@ class IdleSlaves(object):
     def __contains__(self, slave):
         return slave in self._all_slaves
 
+    def __len__(self):
+        return len(self._all_slaves)
+
 # vim: et sw=4 sts=4
