@@ -1,24 +1,17 @@
 # Mrs
 # Copyright 2008-2012 Brigham Young University
 #
-# This file is part of Mrs.
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
 #
-# Mrs is free software: you can redistribute it and/or modify it under the
-# terms of the GNU General Public License as published by the Free Software
-# Foundation, either version 3 of the License, or (at your option) any later
-# version.
+#     http://www.apache.org/licenses/LICENSE-2.0
 #
-# Mrs is distributed in the hope that it will be useful, but WITHOUT ANY
-# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
-# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
-# details.
-#
-# You should have received a copy of the GNU General Public License along with
-# Mrs.  If not, see <http://www.gnu.org/licenses/>.
-#
-# Inquiries regarding any further use of Mrs, please contact the Copyright
-# Licensing Office, Brigham Young University, 3760 HBLL, Provo, UT 84602,
-# (801) 422-9339 or 422-3821, e-mail copyright@byu.edu.
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 """Mrs Runner"""
 
@@ -32,6 +25,7 @@ import time
 from . import job
 from . import peons
 from . import computed_data
+from . import serializers
 from . import util
 from . import worker
 
@@ -71,8 +65,6 @@ class BaseRunner(object):
         close_requests: set of Datasets requested to be closed by the job
             process (but which cannot be closed until their dependents are
             computed)
-        computing_datasets: set of Datasets that are being or need to be
-            computed
         data_dependents: maps a dataset id to a deque listing datasets that
             cannot start until it has finished
         datasets: maps a dataset id to the corresponding Dataset object
@@ -95,9 +87,7 @@ class BaseRunner(object):
                     self.read_worker_pipe)
 
         self.datasets = {}
-        self.computing_datasets = set()
         self.data_dependents = collections.defaultdict(collections.deque)
-        self.close_requests = set()
 
     def read_job_conn(self):
         try:
@@ -107,17 +97,19 @@ class BaseRunner(object):
 
         if isinstance(message, job.DatasetSubmission):
             ds = message.dataset
+            # Fix the breaking of serializers caused by pickling.
+            if ds.serializers is not None:
+                ds.serializers = serializers.from_names(ds.serializers,
+                        self.program_class)
             self.datasets[ds.id] = ds
             input_id = getattr(ds, 'input_id', None)
             if input_id:
                 self.data_dependents[input_id].append(ds.id)
             if isinstance(ds, computed_data.ComputedData):
-                self.computing_datasets.add(ds)
                 self.compute_dataset(ds)
         elif isinstance(message, job.CloseDataset):
-            self.close_requests.add(message.dataset_id)
-            self.try_to_close_dataset(message.dataset_id)
-            self.try_to_remove_dataset(message.dataset_id)
+            ds = self.datasets[message.dataset_id]
+            self.close_dataset(ds)
         elif isinstance(message, job.JobDone):
             if message.exitcode != 0:
                 logger.critical('Job execution failed.')
@@ -138,19 +130,23 @@ class BaseRunner(object):
         """Called when a dataset's computation is finished."""
 
         dataset.computation_done()
-        self.computing_datasets.remove(dataset)
 
         # Check whether any datasets can be closed as a result of the newly
         # completed computation.
-        if dataset.input_id:
-            self.try_to_close_dataset(dataset.input_id)
-        self.try_to_close_dataset(dataset.id)
+        #if dataset.input_id:
+        #    self.try_to_close_dataset(dataset.input_id)
+        #self.try_to_close_dataset(dataset.id)
 
         self.send_dataset_response(dataset)
-        self.try_to_remove_dataset(dataset.id)
+
+        # TODO: preface with `if not self.opts.checkpointing:` or similar.
+        input_id = getattr(dataset, 'input_id', None)
+        # Completing computation decrements the refcount of the input dataset.
+        self.data_dependents[input_id].remove(dataset.id)
+        self.try_to_remove_recursive(input_id)
 
     def send_dataset_response(self, dataset):
-        if not (dataset.closed or dataset.id in self.close_requests):
+        if not dataset.closed:
             for bucket in dataset[:, :]:
                 if len(bucket) or bucket.url:
                     response = job.BucketReady(dataset.id, bucket)
@@ -160,63 +156,79 @@ class BaseRunner(object):
 
     def close_dataset(self, dataset):
         """Close the given dataset, performing any necessary cleanup."""
-        logger.debug('Closing dataset: %s' % dataset.id)
         dataset.close()
+        # Since closing decrements the refcount, try to remove the dataset
+        # and any eligible parents.
+        self.try_to_remove_recursive(dataset.id)
 
-    def try_to_close_dataset(self, dataset_id):
-        """Try to close the given dataset and remove its parent."""
-        if dataset_id not in self.close_requests:
-            return
-        ds = self.datasets[dataset_id]
-        # Skip datasets that are currently being computed.
-        if ds in self.computing_datasets:
-            return
-        # Skip datasets that haven't been computed yet.
-        if getattr(ds, 'computing', False):
-            return
-        # Bail out if any dependent dataset still needs to be computed.
-        deplist = self.data_dependents[dataset_id]
-        for dependent_id in deplist:
-            dependent_ds = self.datasets[dependent_id]
-            if getattr(dependent_ds, 'computing', False):
-                return
+    def try_to_remove_recursive(self, dataset_id):
+        """Try to remove the given dataset and any eligible ancestors.
 
-        self.close_dataset(ds)
-        self.close_requests.remove(dataset_id)
+        Dataset removal is conceptually acyclic garbage collection, and we use
+        reference counts to determine whether a dataset can be removed.  There
+        are two types of references.  First, if a dataset is open, then the
+        run method is holding a reference.  Second, if a dataset is the input
+        to some other "dependent" dataset, than that other dataset holds a
+        reference.
 
-        input_id = getattr(ds, 'input_id', None)
-        if input_id:
-            self.data_dependents[input_id].remove(dataset_id)
-            self.try_to_remove_dataset(input_id)
+        Removal is recursive because removing a dataset decrements the
+        refcount of its input dataset.
+        """
+        candidate_ids = [dataset_id]
+        while candidate_ids:
+            cand_id = candidate_ids.pop()
+            new_candidates = self.try_to_remove_single(cand_id)
+            if new_candidates:
+                candidate_ids.extend(new_candidates)
 
-    def try_to_remove_dataset(self, dataset_id):
+    def try_to_remove_single(self, dataset_id):
         """Try to remove the given dataset.
 
-        If a dataset is not closed or if any of its direct dependents are not
-        closed, then it will not be removed.
-        """
-        if self.can_remove_dataset(dataset_id):
-            self.remove_dataset(self.datasets[dataset_id])
-
-    def can_remove_dataset(self, dataset_id):
-        """Determine whether the given dataset can be safely removed.
-
-        If a dataset is not closed or if any of its direct dependents are not
-        closed, then it will not be removed.
+        Returns a list of dataset ids which may possibly be eligible for
+        removal if this dataset was removed, or None otherwise.
         """
         ds = self.datasets.get(dataset_id, None)
-        if ds is None:
-            return False
-        if not ds.closed:
+        if not ds or not self.can_remove_dataset(ds):
+            return
+
+        input_id = getattr(ds, 'input_id', None)
+        self.remove_dataset(ds)
+        # Decrement the refcount on the input dataset.
+        if input_id:
+            input_dependents = self.data_dependents[input_id]
+            if dataset_id in input_dependents:
+                input_dependents.remove(dataset_id)
+                return [input_id]
+            else:
+                return
+        else:
+            return
+
+    def can_remove_dataset(self, dataset):
+        """Determine whether the given dataset can be safely removed."""
+
+        # If the dataset is open, then its refcount is non-zero.
+        if not dataset.closed:
             return False
 
-        deplist = self.data_dependents[dataset_id]
+        # If the dataset has dependents, then its refcount is non-zero.
+        deplist = self.data_dependents[dataset.id]
         if deplist:
             return False
 
+        # Skip permanent datasets that are not yet computed.
+        if dataset.permanent and getattr(dataset, 'computing', False):
+            return False
+
         return True
 
     def remove_dataset(self, dataset):
+        """Remove the given dataset.
+
+        The dataset and its related structures are cleared from RAM.  If the
+        dataset is non-permenant (any files are purely temporary), then its
+        files are also deleted.
+        """
         logger.info('Removing dataset: %s' % dataset.id)
         if dataset.permanent:
             dataset.clear()
@@ -240,7 +252,6 @@ class TaskRunner(BaseRunner):
         pending_datasets: set of datasets that are not yet runnable
         runnable_datasets: deque of datasets that are ready to run but have
             not yet been split into tasks
-        ready_tasks: list of (dataset, source) pairs
         tasklists: map from a dataset id to the corresponding TaskList
         forward_links: map from a dataset id to a set of ids representing
             the transitive closure of datasets that backlink to it
@@ -306,21 +317,18 @@ class TaskRunner(BaseRunner):
 
         super(TaskRunner, self).close_dataset(dataset)
 
-    def can_remove_dataset(self, dataset_id):
-        """Determine whether the given dataset can be safely removed.
+    def can_remove_dataset(self, dataset):
+        """Determine whether the given dataset can be safely removed."""
 
-        If a dataset is not closed or if any of its direct dependents are not
-        closed, then it will not be removed.
-        """
-        if not super(TaskRunner, self).can_remove_dataset(dataset_id):
+        if not super(TaskRunner, self).can_remove_dataset(dataset):
             return False
-        if dataset_id not in self.forward_links:
+        if dataset.id not in self.forward_links:
             return True
 
-        if self.forward_links[dataset_id]:
+        if self.forward_links[dataset.id]:
             return False
         else:
-            del self.forward_links[dataset_id]
+            del self.forward_links[dataset.id]
             return True
 
     def next_task(self):
@@ -396,30 +404,35 @@ class TaskRunner(BaseRunner):
     def task_done(self, dataset_id, task_index, outurls, backlinked=False):
         """Report that the given source of the given dataset is computed.
 
+        Returns False if the task has already been reported as done (duplicate
+        work).
+
         Arguments:
             dataset_id: string
             task_index: integer id of the task that produced the data
             outurls: list of (number, string) pairs representing the split and
                 url of the outputs.
         """
+        tasklist = self.tasklists[dataset_id]
+        if tasklist.is_task_done(task_index):
+            return False
+
         if not backlinked:
             self.task_counter += 1
-        tasklist = self.tasklists[dataset_id]
         tasklist.task_done(task_index)
 
         dataset = self.datasets[dataset_id]
-        if not dataset.closed:
-            for split, url in outurls:
-                bucket = dataset[task_index, split]
-                bucket.url = url
-                if dataset_id not in self.close_requests:
-                    response = job.BucketReady(dataset_id, bucket)
-                    self.job_conn.send(response)
-            if tasklist.time_to_report_progress():
-                response = job.ProgressUpdate(dataset_id,
-                        tasklist.fraction_complete())
+        for split, url in outurls:
+            bucket = dataset[task_index, split]
+            bucket.url = url
+            if not dataset.closed:
+                response = job.BucketReady(dataset_id, bucket)
                 self.job_conn.send(response)
-            dataset.notify_urls_known()
+        if tasklist.time_to_report_progress():
+            response = job.ProgressUpdate(dataset_id,
+                    tasklist.fraction_complete())
+            self.job_conn.send(response)
+        dataset.notify_urls_known()
 
         if tasklist.complete():
             self.dataset_done(dataset)
@@ -436,12 +449,17 @@ class TaskRunner(BaseRunner):
                             backlinked=True)
 
         self._wakeup_dependents(dataset_id)
+        return True
 
     def task_lost(self, dataset_id, task_index):
         """Report that a task was lost (e.g., to a dead slave)."""
         tasklist = self.tasklists.get(dataset_id)
         if tasklist is not None:
-            tasklist.push(task_index)
+            failure_count = tasklist.task_failed(task_index)
+            if failure_count > self.opts.mrs__max_failures:
+                msg = ("Task failed too many times (see slave logs "
+                        "for details): (%s, %s)" % (dataset_id, task_index))
+                raise RuntimeError(msg)
 
     def dataset_done(self, dataset):
         self.runnable_datasets.remove(dataset)
@@ -460,7 +478,7 @@ class TaskRunner(BaseRunner):
             fraction_complete = 1
 
         if fraction_complete < 1:
-            if fraction_complete < ds.blocking_percent:
+            if fraction_complete < ds.blocking_ratio:
                 return
             if self.available_tasks() >= self.available_workers():
                 return
@@ -531,15 +549,17 @@ class TaskRunner(BaseRunner):
 
         self.timing_stats()
 
-        print('Runnable datasets:', (', '.join(ds.id
-                for ds in self.runnable_datasets)), file=sys.stderr)
-        print('Pending datasets:', (', '.join(ds.id
-                for ds in self.pending_datasets)), file=sys.stderr)
+        print('Runnable datasets:',
+                (', '.join(ds.id for ds in self.runnable_datasets)),
+                file=sys.stderr)
+        print('Pending datasets:',
+                (', '.join(ds.id for ds in self.pending_datasets)),
+                file=sys.stderr)
         print('Ready tasks:', file=sys.stderr)
         for ds in self.runnable_datasets:
-            if ds in self.tasklists:
-                sources = (str(t[1]) for t in self.tasklists[ds])
-                print('    %s:' % ds.id, ', '.join(sources))
+            if ds.id in self.tasklists:
+                taskids = (str(taskid) for _, taskid in self.tasklists[ds.id])
+                print('    %s:' % ds.id, ', '.join(taskids), file=sys.stderr)
 
 
 class TaskList(object):
@@ -552,7 +572,9 @@ class TaskList(object):
         self._ready_tasks = collections.deque()
         self._tasks_made = False
         self._async_incomplete = None
+        self._num_tasks = 0
         self._last_progress_report = 0.0
+        self._failures = collections.defaultdict(int)
 
     def make_tasks(self, done_tasks, backlink_tasks, incomplete_sources):
         """Generate tasks for the given dataset, adding them to ready_tasks.
@@ -576,6 +598,7 @@ class TaskList(object):
                         self._ready_tasks.append(task_index)
                         self._remaining_tasks.add(task_index)
                         break
+        self._num_tasks = len(self._remaining_tasks)
         self._tasks_made = True
 
     def async_incomplete(self):
@@ -591,7 +614,7 @@ class TaskList(object):
     def fraction_complete(self):
         """Returns the fraction of datasets that have been computed."""
         if self._tasks_made:
-            return 1 - len(self._remaining_tasks) / self.dataset.ntasks
+            return 1 - len(self._remaining_tasks) / self._num_tasks
         else:
             return 0
 
@@ -625,12 +648,15 @@ class TaskList(object):
         except IndexError:
             return None
 
-    def push(self, task_index):
-        """Push back a task.
+    def task_failed(self, task_index):
+        """Push back a failed task.
 
-        Called if, for example, a Task is aborted.
+        Called if, for example, a Task is aborted.  Returns the number of
+        times that this particular task has failed.
         """
-        self._ready_tasks.appendleft(task_index)
+        self._ready_tasks.append(task_index)
+        self._failures[task_index] += 1
+        return self._failures[task_index]
 
     def __iter__(self):
         """Iterate over tasks ((dataset_id, task_index) pairs)."""
@@ -677,7 +703,7 @@ class MockParallelRunner(TaskRunner, worker.WorkerManager):
         self.schedule()
 
     def worker_failure(self, r):
-        """Called when a worker sends a WorkerSuccess."""
+        """Called when a worker sends a WorkerFailure."""
         raise RuntimeError('Task failed')
 
 # vim: et sw=4 sts=4
